import hashlib
import hmac
import io
import json
import time
import unittest
import uuid
from collections import OrderedDict

import bcrypt
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.utils import timezone
from rest_framework.test import APIClient

from rest_framework_signature.settings import auth_settings


def get_hours_in_milliseconds(hours):
    return hours * 60 * 60 * 1000


def get_nonce(timestamp, url, secret_access_key, body=None):
    m = hmac.new(secret_access_key.encode(), digestmod=hashlib.sha1)
    m.update(str(timestamp).encode('utf-8'))
    m.update(url.encode('utf-8'))
    if body:
        sorted_body = sort_body(body)
        sorted_json = json.dumps(sorted_body, ensure_ascii=False).replace(' ', '')
        m.update(sorted_json.encode('utf-8'))
    return m.hexdigest()


def get_timestamp_milliseconds(dt=None):
    if not dt:
        dt = timezone.now()
    # grab time since epoch milliseconds
    return int(time.mktime(dt.timetuple())*1000 + dt.microsecond/1000)


def sort_body(data):
    res = OrderedDict()
    for k, v in sorted(data.items()):
        # everything will be in unicode unless we've converted it and went to the next level of depth
        if isinstance(v, str):
            try:
                parsed_data = json.loads(v)
            except ValueError:
                parsed_data = v
            if isinstance(parsed_data, dict):
                res[k] = sort_body(parsed_data)
            elif isinstance(parsed_data, list):
                res[k] = [sort_body(list_item) for list_item in parsed_data]
            else:
                res[k] = v
        else:
            if isinstance(v, dict):
                res[k] = sort_body(v)
            elif isinstance(v, list):
                res[k] = [sort_body(list_item) for list_item in v]
            # if its a file just continue
            elif isinstance(v, io.TextIOWrapper) or isinstance(v, InMemoryUploadedFile):
                continue
            else:
                res[k] = v
    return res


def check_valid_reset_token(reset_token, user):
    # check if the reset_token has expired
    expire_time = get_timestamp_milliseconds(user.password_reset_token_expires)
    current_time = get_timestamp_milliseconds()
    is_expired = expire_time < current_time

    if user.password_reset_token != reset_token or is_expired:
        return False

    return True


def generate_email_address():
    return '{0}@{1}.com'.format(str(uuid.uuid4())[:10], str(uuid.uuid4())[:5])


class RestFrameworkSignatureTestClass(unittest.TestCase):

    user_model = auth_settings.get_user_document()
    application_model = auth_settings.get_application_document()
    auth_token_model = auth_settings.get_auth_token_document()

    def get_headers(self, url, body=None):
        headers = self.get_headers_without_auth(url, body)
        headers['HTTP_AUTHORIZATION'] = 'Token {0}'.format(self.token.key)
        return headers

    def get_headers_without_auth(self, url, body=None):
        timestamp = str(get_timestamp_milliseconds())
        nonce = get_nonce(str(timestamp), url, self.device_token.secret_access_key, body=body)

        headers = {
            auth_settings.TIMESTAMP_HEADER: str(timestamp),
            auth_settings.NONCE_HEADER: nonce,
            auth_settings.API_KEY_HEADER: self.device_token.access_key
        }
        return headers

    def setUp(self):
        # create client to access api endpoints
        self.api_client = APIClient()

        # create a user to use to authenticate against
        username = generate_email_address()
        salt = bcrypt.gensalt()
        m = hashlib.sha1()
        m.update('pass1234'.encode('utf-8'))
        sha1_password = m.hexdigest()
        m = hashlib.sha1()
        m.update(sha1_password.encode('utf-8'))
        m.update(salt)
        password = m.hexdigest()
        test_user = self.user_model(
            username=username,
            password=password,
            salt=salt
        )
        test_user.save()

        # create an authentication token
        token, created = self.auth_token_model.objects.get_or_create(user=test_user)

        # create a signature DeviceToken to hash our requests
        access_key = str(uuid.uuid4())[:10]
        secret_access_key = str(uuid.uuid4())[:20]
        device_token = self.application_model(
            name='test-app',
            access_key=access_key,
            secret_access_key=secret_access_key
        )
        device_token.save()

        self.token = token
        self.user = test_user
        self.sha1_password = sha1_password
        self.device_token = device_token

    def tearDown(self):
        pass

    class MockRequestsResponse:
        def __init__(self, status_code, text, mock_json_cb=None):
            self.status_code = status_code
            self.text = text
            self.mock_json_cb = mock_json_cb

        def json(self):
            return self.mock_json_cb



class ErrorMessages:
    # security
    NO_PASSWORD = 'Please enter a password.'
    INVALID_EMAIL = 'Email address is invalid.'
    NO_USERNAME = 'Please enter an email address.'
    NOT_UNIQUE_EMAIL = 'This email is already registered. Sign in.'
    NO_FB_AUTH_TOKEN = 'Facebook token missing.'
    INVALID_FB_AUTH_TOKEN = 'Invalid or expired Facebook token.'
    MISSING_USERNAME = 'Please enter a valid e-mail address.'
    INVALID_USERNAME = 'E-mail address does not exist in the system.'
    INVALID_RESET_PASSWORD_TOKEN = 'Reset password link is expired or invalid.'
    ACCOUNT_DISABLED = 'User account is disabled.'
    INVALID_CREDENTIALS = 'Unable to log in with provided credentials.'
    MISSING_USERNAME_OR_PASSWORD = 'Must include username and password.'
    TOO_MANY_INCORRECT_LOGIN_ATTEMPTS = 'Too many login attempts. Please try again in 20 minutes.'
    PERMISSION_DENIED = 'You do not have the permission to perform this action.'
    INVALID_API_KEY = 'Error 1001 API key not found'
    MISSING_API_KEY = 'Error 1002 No API key provided'
    REPLAY_ERROR = 'Error 1003'
    INVALID_NONCE = 'Error 1004 Provided nonce does not match computed nonce'
    MISSING_NONCE = 'Error 1005 No Nonce provided'
    NO_SSO_TOKEN_CLASSES_DEFINED = 'Misconfigured SSO Token Classes'
    NO_SSO_TOKEN_FOUND = 'No SSO token found - please regenerate'
    API_KEY_NOT_AUTHORIZED_FOR_ENDPOINT = 'ApiKey is not authorized for endpoint {0}'




import binascii
from datetime import timedelta
import hashlib
import os

import bcrypt
from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone
from mongoengine.errors import DoesNotExist
from rest_framework_signature.helpers import check_valid_reset_token
from rest_framework_signature.serializers import AuthTokenSerializer, SSOTokenSerializer
from rest_framework_signature.settings import auth_settings
from rest_framework import status, authentication
from rest_framework.authtoken.views import ObtainAuthToken
from rest_framework.response import Response
from rest_framework.views import APIView

from rest_framework_signature.errors import ErrorMessages


class DeleteAuthToken(APIView):
    def post(self, request):
        auth_token_model = auth_settings.get_auth_token_document()
        auth = authentication.get_authorization_header(request).split()

        try:
            token = auth_token_model.objects.get(user=request.user, key=auth[1].decode('utf-8'))
        except (DoesNotExist, ObjectDoesNotExist):
            return Response(status=status.HTTP_400_BAD_REQUEST)
        token.delete()

        if auth_settings.DATABASE_ENGINE == 'mongo':
            # have to cascade_save to ensure parent documents are deleted -- due to inheritance in mongoengine
            token.cascade_save()
        return Response(status=status.HTTP_204_NO_CONTENT)


class GetAuthToken(ObtainAuthToken):
    auth_token_model = auth_settings.get_auth_token_document()
    model = auth_token_model
    user_model = auth_settings.get_user_document()
    # set authentication classes to empty as this needs to be open for them to obtain their auth token
    authentication_classes = ()
    serializer_class = AuthTokenSerializer

    def post(self, request):

        serializer = self.serializer_class(data=request.data)
        if serializer.is_valid():
            user = serializer.validated_data['user']
            token, created = self.model.objects.get_or_create(user=user, auth_type='password')
            if not created:
                token.updated = timezone.now()
                token.save()

            user.failed_login_attempts = 0
            user.save()
            response = {'token': token.key, 'userId': serializer.validated_data['user'].id}
            return Response(response, content_type='application/json', status=status.HTTP_200_OK)

        # increase incorrect login attempts
        username = request.data.get('username', None)
        if username:
            try:
                user = self.user_model.objects.get(username=username)
                if user.failed_login_attempts is None:
                    user.failed_login_attempts = 0
                user.failed_login_attempts += 1
                user.last_failed_login = timezone.now()
                user.save()
                if user.failed_login_attempts >= auth_settings.FAILED_LOGIN_RETRY_ATTEMPTS:
                    response = {'error_message': ErrorMessages.TOO_MANY_INCORRECT_LOGIN_ATTEMPTS}
                    return Response(response, content_type='application/json', status=status.HTTP_400_BAD_REQUEST)
            except (DoesNotExist, ObjectDoesNotExist):
                pass
        response = {'error_message': serializer.friendly_error_message}
        return Response(response, content_type='application/json', status=status.HTTP_400_BAD_REQUEST)


class GetAuthTokenSSO(ObtainAuthToken):
    auth_token_model = auth_settings.get_auth_token_document()
    model = auth_token_model
    user_model = auth_settings.get_user_document()
    # set authentication classes to empty as this needs to be open for them to obtain their auth token
    authentication_classes = ()
    serializer_class = SSOTokenSerializer

    def post(self, request):

        serializer = self.serializer_class(data=request.data)
        if serializer.is_valid():
            user = serializer.validated_data['user']
            token, created = self.model.objects.get_or_create(user=user, auth_type='sso')
            if not created:
                token.updated = timezone.now()
                token.save()

            user.failed_login_attempts = 0
            user.save()
            response = {'token': token.key, 'userId': serializer.validated_data['user'].id}
            return Response(response, content_type='application/json', status=status.HTTP_200_OK)

        # todo: block them if too many sso attempts?
        response = {'error_message': ErrorMessages.INVALID_CREDENTIALS}
        return Response(response, content_type='application/json', status=status.HTTP_400_BAD_REQUEST)


class ResetPassword(APIView):
    user_model = auth_settings.get_user_document()

    def post(self, request):

        username = request.data.get('username', None)
        if not username:
            response = {'error_message': ErrorMessages.MISSING_USERNAME}
            return Response(response, content_type='application/json', status=status.HTTP_400_BAD_REQUEST)

        try:
            user = self.user_model.objects.get(username=username)
        except (DoesNotExist, ObjectDoesNotExist):
            response = {'error_message': ErrorMessages.INVALID_USERNAME}
            return Response(response, content_type='application/json', status=status.HTTP_400_BAD_REQUEST)

        # create verification token
        user.password_reset_token = binascii.hexlify(os.urandom(22)).decode()
        user.password_reset_token_created = timezone.now()
        user.password_reset_token_expires = timezone.now() + timedelta(hours=auth_settings.RESET_PASSWORD_TOKEN_EXPIRATION)
        user.save()

        # send email
        # send email with password_reset_token link
        response = {'reset_password_token': user.password_reset_token, 'user_email': user.email}
        return Response(response, status=status.HTTP_200_OK)


class CheckPasswordResetLink(APIView):
    user_model = auth_settings.get_user_document()

    def post(self, request):
        reset_token = request.data.get('reset_token', None)

        if not reset_token:
            response = {'error_message': ErrorMessages.INVALID_RESET_PASSWORD_TOKEN}
            return Response(response, content_type='application/json', status=status.HTTP_400_BAD_REQUEST)

        try:
            user = self.user_model.objects.get(password_reset_token=reset_token)
        except (DoesNotExist, ObjectDoesNotExist):
            response = {'error_message': ErrorMessages.INVALID_RESET_PASSWORD_TOKEN}
            return Response(response, content_type='application/json', status=status.HTTP_400_BAD_REQUEST)

        valid_reset_token = check_valid_reset_token(reset_token, user)
        if not valid_reset_token:
            response = {'error_message': ErrorMessages.INVALID_RESET_PASSWORD_TOKEN}
            return Response(response, content_type='application/json', status=status.HTTP_400_BAD_REQUEST)

        response = {'success': True}
        return Response(response, status=status.HTTP_200_OK)


class SubmitNewPassword(APIView):
    user_model = auth_settings.get_user_document()

    def post(self, request):

        reset_token = request.data.get('reset_token', None)

        if not reset_token:
            response = {'error_message': ErrorMessages.INVALID_RESET_PASSWORD_TOKEN}
            return Response(response, content_type='application/json', status=status.HTTP_400_BAD_REQUEST)

        try:
            user = self.user_model.objects.get(password_reset_token=reset_token)
        except (DoesNotExist, ObjectDoesNotExist):
            response = {'error_message': ErrorMessages.INVALID_RESET_PASSWORD_TOKEN}
            return Response(response, content_type='application/json', status=status.HTTP_400_BAD_REQUEST)

        valid_reset_token = check_valid_reset_token(reset_token, user)
        if not valid_reset_token:
            response = {'error_message': ErrorMessages.INVALID_RESET_PASSWORD_TOKEN}
            return Response(response, content_type='application/json', status=status.HTTP_400_BAD_REQUEST)

        password = request.data.get('password', None)
        if not password:
            response = {'error_message': ErrorMessages.NO_PASSWORD}
            return Response(response, content_type='application/json', status=status.HTTP_400_BAD_REQUEST)

        m = hashlib.sha1()
        if user.salt is None:
            user.salt = bcrypt.gensalt()
        m.update(password.encode('utf-8'))
        m.update(user.salt.encode('utf-8'))
        user.password = m.hexdigest()
        user.password_reset_ip = request.META.get('REMOTE_ADDR', None)
        user.password_reset_token = None
        user.save()

        response = {'success': True}
        return Response(response, status=status.HTTP_200_OK)


class Ping(APIView):
    user_model = auth_settings.get_user_document()
    def get(self, request):
        response = {'success': True}
        return Response(response, status=status.HTTP_200_OK)


check_password_reset_link = CheckPasswordResetLink.as_view()
delete_auth_token = DeleteAuthToken.as_view()
obtain_auth_token = GetAuthToken.as_view()
obtain_auth_token_sso = GetAuthTokenSSO.as_view()
reset_password = ResetPassword.as_view()
submit_new_password = SubmitNewPassword.as_view()
ping = Ping.as_view()

